/*
This is the master file used on my electric scooter project
Functionality: 
    Reads two battery temp sensors
    Attach Hall sensor as interrupt
    Saves interrupt to EEPROM
    Calculates mileage and speed
    Reads LiPo battery voltage
    Outputs to an LCD screen via I2c
            

by: Matthew Daneil
Written: 2019-December
Contact: mrd89@cornell.edu

*/

// included packages
#include <EEPROM.h> // EEPROM memory package
#include <TimerOne.h> // timer interrupts
#include <Wire.h> // ic2 
#include <LiquidCrystal_I2C.h> //lcd driver


// defining temperature/battery analog in pins
#define TempPin1 0
#define TempPin2 1
#define voltagePin 2




// global varibales defined

// init temp readings
int temp1 = (int)(500 / 1024.0 * analogRead(TempPin1));
int temp2 = (int)(500 / 1024.0 * analogRead(TempPin2));

// mileage of system, voltage, velocity;
double batteryVoltage, velocity; 
long double mileage;
double milesLeft;

// array of velocity (used for averaging)
double velocity_array[4];



// read and set initial mileage from EEPROM //
long four = EEPROM.read(0);
long three = EEPROM.read(1);
long two = EEPROM.read(2);
long one = EEPROM.read(3);

//Return the recomposed long by using bitshift.
long prev_mileage = ((four << 0) & 0xFF) + ((three << 8) & 0xFFFF) + ((two << 16) & 0xFFFFFF) + ((one << 24) & 0xFFFFFFFF);


// int output to LCD
int lcdvoltage; 

// LCD setup with A4 A5 as i2c pins
// These pins depend on arduino / used pins
LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE); 


void setup() {

  lcd.begin(20, 4); // begins 20 x4 lcd screen
  lcd.clear(); // reset LCD screen


  // setup interrupt on pin 2 (hall sensor)
  pinMode(2, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(2), mileageCounter, FALLING);



  // timer interrupt happening at 4Hz
  Timer1.initialize(250000); // run about every .25 second

  // calls the speedCheck function at Timer1 interrupt
  Timer1.attachInterrupt(speedCheck); 



}



// main control loop
void loop() {

  // check temperature of batteries
  temp1 = (int)(500 / 1024.0 * analogRead(TempPin1));
  temp2 = (int)(500 / 1024.0 * analogRead(TempPin2));


  // sum velocity array. Each velocity array is 1/4 the actual to reduce compute time
  velocity =   (velocity_array[0] + velocity_array[1] + velocity_array[2] + velocity_array[3]);

  // read battery voltage
  batteryVoltage = analogRead(voltagePin) * 10;

// calculate total miles traveled from EEPROM
  mileage = (double)(  prev_mileage * .2 * PI * 0.000621371192 ); // in miles

  // Displays updated info to LCD
  displayLCD();

  // update LCD at 2HZ
  delay(500);

}




// Calculates Speed
void speedCheck() {

  // EEPROM read 32 bit int
  long four = EEPROM.read(0);
  long three = EEPROM.read(1);
  long two = EEPROM.read(2);
  long one = EEPROM.read(3);

  //Return the recomposed long by using bitshift.
  long tempi = ((four << 0) & 0xFF) + ((three << 8) & 0xFFFF) + ((two << 16) & 0xFFFFFF) + ((one << 24) & 0xFFFFFFFF);

  int dx = (int)(tempi - prev_mileage);

  // update velocity array
  velocity_array[0] = velocity_array[1];
  velocity_array[1] = velocity_array[2];
  velocity_array[2] = velocity_array[3];

  // conver change in reading to velocity /4 to reduce compute time
  velocity_array[3] = dx * 1.3055085;

  // update previous mileage
  prev_mileage = tempi;
}


// updates the mileage to +1 per interrupt
void mileageCounter() {
  long temp = EEPROMRead(0);
  temp ++;
  EEPROMWrite(0, temp);
}




// display the LCD screen using global variables
void displayLCD() {

  //clear LCD memory and initiate curser to 0,0 (upper left corner)
  lcd.clear();
  lcd.setCursor(0, 0);

  // Speed display
  lcd.print("SPEED:");
  lcd.setCursor(6, 0);
  lcd.print(velocity);



  //Temp 1 Display
  lcd.setCursor(13, 0);
  lcd.print("T1=");
  lcd.setCursor(16, 0);
  lcd.print(temp1);
  lcd.setCursor(19, 0);
  lcd.print("F");

  //Temp 2 Display
  lcd.setCursor(13, 1);
  lcd.print("T2=");
  lcd.setCursor(16, 1);
  lcd.print(temp2);
  lcd.setCursor(19, 1);
  lcd.print("F");

  // mileage display
  lcd.setCursor(0, 1);
  lcd.print("MILES:");
  lcd.print((double)mileage);


  // Battery voltage and percent
  lcd.setCursor(0, 2);
  lcd.print("VOLTAGE:");
  lcdvoltage = (int)(batteryVoltage * 5 / 1023) ;
  lcd.print(lcdvoltage);
  lcd.print(" ");
  lcd.setCursor(16, 2);
  lcd.print((int) ((lcdvoltage - 40) * 10));
  lcd.setCursor(19, 2);
  lcd.print("%");

}



// EEPROM write helper function
void EEPROMWrite(int address, long value)
{
  //Decomposition from a long to 4 bytes by using bitshift.
  //One = Most significant -> Four = Least significant byte
  byte four = (value & 0xFF);
  byte three = ((value >> 8) & 0xFF);
  byte two = ((value >> 16) & 0xFF);
  byte one = ((value >> 24) & 0xFF);

  //Write the long onto four bytes
  EEPROM.write(address, four);
  EEPROM.write(address + 1, three);
  EEPROM.write(address + 2, two);
  EEPROM.write(address + 3, one);
}


long EEPROMRead(int address)
{
  //Read the 4 bytes from memory
  long four = EEPROM.read(address);
  long three = EEPROM.read(address + 1);
  long two = EEPROM.read(address + 2);
  long one = EEPROM.read(address + 3);

  //Return the long via bitshifting
  return ((four << 0) & 0xFF) + ((three << 8) & 0xFFFF) + ((two << 16) & 0xFFFFFF) + ((one << 24) & 0xFFFFFFFF);
}
